/*
 * Copyright (C) 1994-1996, Index Data I/S 
 * All rights reserved.
 * Sebastian Hammer, Adam Dickmeiss
 *
 * keychars.c,v
 * Revision 1.8  1996/06/04  10:20:58  quinn
 * Added tolower() in default charmap.
 *
 * Revision 1.7  1996/06/04  10:17:41  adam
 * Added const to map_chrs_output.
 *
 * Revision 1.6  1996/06/04  08:32:05  quinn
 * Moved default keymap to keychars.c
 *
 * Revision 1.5  1996/06/04  08:20:16  quinn
 * Smallish
 *
 * Revision 1.4  1996/06/04  07:58:06  quinn
 * Fixed return mapping.
 *
 * Revision 1.3  1996/06/04  07:54:56  quinn
 * Added output-map.
 *
 * Revision 1.2  1996/06/03  10:15:25  quinn
 * Various character-mapping.
 *
 * Revision 1.1  1996/05/31  09:06:59  quinn
 * Work on character-set handling
 *
 *
 */

#include <alexutil.h>
#include <charmap.h>
#include <ctype.h>

#include "index.h"

static chrmaptab *maptab = 0; /* global character mapping setup */

void init_charmap(void)
{
    char *name = res_get(common_resource, "charMap");

    if (name)
    {
	if (!(maptab = chr_read_maptab(name)))
	{
	    logf(LOG_WARN, "Failed to read character table");
	    exit(1);
	}
    }
}

char **map_chrs_input(char **from, int len)
{
    if (!maptab) /* no table loaded. Null mapping */
    {
	static char str[2] = {0,0}, *buf[2] = {0,0};

	if (isalnum(**from))
	{
	    str[0] = isupper(**from) ? tolower(**from) : **from;
	    buf[0] = str;
	}
	else
	    buf[0] = (char*) CHR_SPACE;
	(*from)++;
	return buf;
    }
    return chr_map_input(maptab->input, (char**)from, len);
}

const char *map_chrs_output(const char **from)
{
    unsigned char i = (unsigned char) **from;
    static char buf[2] = {0,0};

    (*from)++;
    if (maptab)
	return (const char *) maptab->output[i];
    else
    {
	buf[0] = i;
	return buf;
    }
}
